# pylint: disable=all

import os
from pathlib import Path
import re
import json
from textwrap import dedent
from typing import Dict, TYPE_CHECKING, Literal, Tuple, Union
from dataclasses import dataclass,field

from bs4.element import Tag
from mkdocs.structure.pages import Page

from mkdocs_addresses import path_manager



if TYPE_CHECKING:
    from .addresses_plugin import AddressAddressesPlugin




#-----------------------------------------------------------
#                  Data extraction tools
#-----------------------------------------------------------



@dataclass
class Content:
    url: str
    target: str

    def to_tests(self):
        return f'''Content(
                    url = { self.url !r},
                    target = { self.target !r},
                ),'''


@dataclass
class Location:
    file_uri: str
    html: str
    data: Tuple[Content,Content]
    md: str = '...'

    @classmethod
    def default(cls, uri:str, html:str):
        return Location(uri, html, (Content('',''), Content('','')))

    def to_json(self):
        return {
            k: getattr(self,k)
            for k in self.__class__.__annotations__     # pylint: disable=no-member
        }

    def to_tests(self):
        sep = "\n                "
        return f'''Location(
                file_uri = { self.file_uri !r},
                html = { self.html !r},
                md = { self.md !r},
                data = ({ ''.join( f"{ sep }{ ct.to_tests() }" for ct in self.data ) }),
            ),'''

@dataclass
class Chunk:
    is_bare:bool
    id:     str
    locations: Dict[str,Tuple[Location,Location]]


    def to_json(self):
        return {
            k: getattr(self,k) if k!='locations' else [
                v.to_json()
                for v in getattr(self,k).values()
            ]
            for k in self.__class__.__annotations__     # pylint: disable=no-member
        }

    def to_tests(self):
        sep = "\n            "
        return f'''\
    Chunk(
        is_bare = { self.is_bare !r},
        id = { self.id !r},
        locations = [{ ''.join( f"{ sep }{ loc.to_tests() }" for loc in self.locations.values() ) }
        ],
    ),\n'''




def dict_to_str(data:dict, name="actual"):
    """ Utility to print a dict to the console """
    return f'\n\n#{ name }:\n' + '{' + ''.join(
        f"\n\t{k!r: <30}: {v!r}," for k,v in sorted(data.items())
    ) + '\n}'


def simplify_ids(id:str):
    return re.sub(r'[ -]other|[ -]?(?<!h)2','',id)



FlowKind = Literal['refs_on_hook','html_content','start_context','end_context']


@dataclass
class LiveDataExtractor:
    """ Tool used to extract various data to build some data sets for the tests """

    dct_misc: Dict[str,Chunk] = field(default_factory=dict)     # not Flow related

    refs_on_hook:  Dict[str,str] = field(default_factory=dict)
    markdown:      Dict[str,str] = field(default_factory=dict)
    html_content:  Dict[str,str] = field(default_factory=dict)
    start_context: Dict[str,str] = field(default_factory=dict)
    end_context:   Dict[str,str] = field(default_factory=dict)


    def clear(self):
        for prop in getattr(LiveDataExtractor,'__annotations__', {}):
            data = getattr(self, prop, None)
            if hasattr(data,'clear'):
                data.clear()

    def refs_snapshot_at(self, hook:str, plugin:'AddressAddressesPlugin'):
        self.refs_on_hook[hook] = plugin.global_handler.references.id_to_data.copy()

    def add_html_and_markdown(self, page:Page, html):
        self.markdown[page.file.src_uri] = page.markdown
        self.html_content[page.file.src_uri] = html

    def add_start_ctx(self, page:Page):
        self.start_context[page.file.src_uri] = page.content

    def add_end_ctx(self, page:Page):
        self.end_context[page.file.src_uri] = page.content



    def flow_dump(self, plugin:'AddressAddressesPlugin'):
        """ To generate the data for all files:

                1. Activate the extractor code: "dev toggle"
                2. Check that everything is fine in yml ad hooks files... (!!)
                3. Set "i_flag = -1" in the mkdocs_dev_hooks.py file
                4. Run successively in the console (until all the flags have been used):
                        dev apply
                        Ctrl+C, start over...
                5. Grab the file "dump/full_flow.json"
        """
        flow     = Flow()
        flow_key = self.get_flow_key(plugin)
        kinds    = 'markdown html_content start_context end_context'.split()
        for kind in kinds:
            flow.push(kind, flow_key, getattr(self,kind))

        kind  = 'refs_on_hook'
        hooks = 'on_nav on_env on_post_build'.split()
        for hook in hooks:
            if hook in self.refs_on_hook:
                flow.push(kind, flow_key, self.refs_on_hook[hook], hook)
        flow.dump()
        self.clear()


    @classmethod
    def get_flow_key(cls, plugin:'AddressAddressesPlugin'):
        docs = plugin.docs_dir
        docs_dir = os.path.basename(docs)
        links_uri, snip_uri = (
            path_manager.to_uri(path_manager.to_os(os.path.relpath(file, docs)))
                for file in (plugin.external_links_file, plugin.dump_snippets_file)
        )
        return (
            f"cache={plugin.activate_cache}, {docs_dir}, vsc={plugin.use_auto_completion}, "
            f"no_headers={plugin.ignore_auto_headers}, { 'fail fast' if plugin.fail_fast else 'skip fail'}, "
            f"ext_links={ links_uri }, snippets_file={ snip_uri }, "
            f"use_dir={plugin.use_directory_urls}"  # at last, because most likely always redundant with docs_dir
        )


    @classmethod
    def get_flow_data(cls, flow_data, plugin:'AddressAddressesPlugin', kind:FlowKind, data:Union[Page,str]):
        key     = cls.get_flow_key(plugin)
        sub_key = data.file.src_uri if isinstance(data,Page) else data
        pif     = flow_data[key]
        paf     = pif[kind]
        value   = paf[sub_key]
        return value


    @classmethod
    def assert_flow(cls, flow_data, plugin:'AddressAddressesPlugin', kind:FlowKind, data:Union[Page,str]):
        expected = cls.get_flow_data(flow_data, plugin, kind, data)
        actual   = data.content if isinstance(data,Page) else plugin.global_handler.references.id_to_data
        assert actual == expected



    #----------------------------------------------------------------------------------
    #                               Non flow routines
    #----------------------------------------------------------------------------------


    def add(self, plugin:'AddressAddressesPlugin', page:Page, id:str, tag:Tag, target:str):
        identifier = simplify_ids(id)
        tag_code = simplify_ids(str(tag))
        if identifier not in self.dct_misc:
            chunk = Chunk('bare-ref' in identifier, identifier, {})
            self.dct_misc[identifier] = chunk
        else:
            chunk = self.dct_misc[identifier]

        loc = chunk.locations.get(page.file.src_uri, Location.default(page.file.src_uri, tag_code))
        chunk.locations[page.file.src_uri] = loc
        use_dir = plugin.use_directory_urls
        loc.data[use_dir].url    = simplify_ids(page.url)
        loc.data[use_dir].target = simplify_ids(target)



    def dump(self, plugin:'AddressAddressesPlugin'):

        with open('dump/dumb.py', mode='w') as f:
            f.write(self.to_tests())
            # f.write('\n#------------------------------------\n#References:\n')

            # for field in 'id_to_data id_to_source source_to_ids snippets'.split():
                # data: dict = getattr(plugin.global_handler.references, field)
                # f.write(dict_to_str(data, field))
            # f.write(str(plugin))
            # f.write(dict_to_str(self.html_content, 'html_content'))
            # f.write(dict_to_str(self.end_context, 'end_context'))



    def to_tests(self):
        return """from .data_classes import *


PAGE_CONTENT_DATA = PoolData([\n""" + "".join(c.to_tests() for c in self.dct_misc.values()) + "\n])\n"





class Flow(dict):
    """
        Flow data structure:
        {
            sources:{
                refs_on_hook:{
                    0: dct,
                    ...
                },
                htmls: {        # holds ANY kind of html code
                    0: dct,
                    ...
                },
                markdown: {
                    0: dct,
                    ...
                }
            },
            flow:{
                flow_key: {
                    refs_on_hook: {
                        hook: "refs_on_hook[i]",
                        ...
                    },
                    markdown: "markdown[i]",
                    html_content: "htmls[i]",
                    start_content: "htmls[i]",
                    end_content: "htmls[i]",
                }
            }
        }
        """

    FILE = Path('dump/full_flow.json')


    def __init__(self, file:Path=None):
        self.file = file or self.FILE
        self.uniques = {}
        flow = self.__load()
        super().__init__(flow)


    def __load(self):
        flow_str = dedent('''
        {
            "sources": {
                "refs_on_hook":{},
                "htmls": {},
                "markdown": {}
            },
            "flow": {}
        }''')
        if self.file.is_file():
            flow_str = self.file.read_text(encoding='utf8') or flow_str

        flow = json.loads(flow_str)
        for source,dct in flow['sources'].items():
            for i,data in dct.items():
                path = f"flow['sources'][{source!r}]['{i}']"
                self.uniques[self.__dct_to_hash(data)] = path
        return flow


    @staticmethod
    def __dct_to_hash(dct:dict):
        return frozenset(dct.items())


    def get_path_for(self, data_kind:str, data:dict):
        key = self.__dct_to_hash(data)
        data_kind = data_kind if data_kind in self['sources'] else 'htmls'

        if key not in self.uniques:
            src = self['sources'][data_kind]
            i = len(src)
            src[i] = data
            path = f"flow['sources'][{ data_kind !r}]['{ i }']"
            self.uniques[key] = path

        return self.uniques[key]


    def push(self, kind:str, flow_key:str, data:dict, hook:str=None):
        if hook is None:
            digging,final_key = ('flow', flow_key), kind
        else:
            digging, final_key = ('flow', flow_key, kind), hook

        dct = self
        for key in digging:
            if key not in dct: dct[key] = {}
            dct = dct[key]
        dct[final_key] = self.get_path_for(kind, data)


    def dump(self):
        ordered_flow = { k:v for k,v in sorted(self['flow'].items()) }
        self['flow'] = ordered_flow
        flow_str = json.dumps(self, indent=2)
        self.file.touch(exist_ok=True)
        self.file.write_text(flow_str, encoding='utf8')


    @classmethod
    def to_test_data(cls, source:Path=None):
        """ Rebuild the original SOURCE_DATA structure for the tests """
        flow = Flow(source or Path("tests/test_mkdocsaddresses/data_sets/full_flow.json"))
        data = { flow_key: {
                kind: eval(data, {'flow':flow}) if isinstance(data,str)
                    else {hook: eval(path, {'flow':flow}) for hook,path in data.items()}
                for kind,data in data_dct.items()
            }
            for flow_key,data_dct in flow['flow'].items()
        }
        return data





from typing import Dict, Literal, Tuple, Type, TYPE_CHECKING

from ._base_handler import RefKind, AutoCompletion, PLUGIN_MARK
from ._vsc_handlers import VscAutoAddress, VscAutoInclude, VscAutoCompletion
from ._no_completion_handler import NoAutoAddress, NoAutoInclude, NoAutoCompletion

if TYPE_CHECKING:
    from mkdocs_addresses.config_plugin import AddressAddressesConfig




IdeKind     = Literal['none','vsc']
NONE,VSC    = 'none','vsc'
IDE_OPTIONS = NONE,VSC


COMPLETION_CLASSES_CONFIG: Dict[IdeKind,Tuple[Type[AutoCompletion],...]]= {
    NONE: (NoAutoCompletion, NoAutoAddress, NoAutoInclude),           # by default...
    VSC:  (VscAutoCompletion, VscAutoAddress, VscAutoInclude),
}



def validate_config_and_get_auto_completion_handler(config:'AddressAddressesConfig') -> Type[AutoCompletion]:
    """
    Build the AutoCompletion instances, for the given type of IDE
    """

    # Handle legacy option
    if config.use_vsc is not None:
        config.ide = VSC if config.use_vsc else NONE
        config.use_vsc = None

    TopKls, KlsAddress, KlsInclude = COMPLETION_CLASSES_CONFIG[config.ide]
    if not config.dump_snippets_file:
        config.dump_snippets_file = TopKls.filename

    handlers = {
        formatter.kind: formatter for formatter in [
            KlsInclude( RefKind.Include, '::'),                     # code inclusions
            KlsAddress( RefKind.Link,    '--'),                     # internal links
            KlsAddress( RefKind.File,    '++'),                     # any internal file
            KlsAddress( RefKind.Img,     '!!', snippet_start='!'),  # images
            KlsAddress( RefKind.Ext,      '', tail_wrap=('[',']'),  # External links
                                              attrs="target=_blank"),
    ]}
    TopKls.setup_handlers(handlers)
    return TopKls



from mkdocs.exceptions import PluginError

# pylint: disable=too-many-ancestors


class AddresserError(PluginError):
    """ Generic mkdocs-addresses exception. """



#------------------------------------------------



class AbortError(AddresserError):
    """ Thrown at the face of mkdocs logger... :p
        (see addresses_auto_log final try/except block)
    """



# class InclusionError(AddresserError):
#     """ A file will result in a code snippet using the same prefix than a previous one """



class DumpOnlyException(AddresserError):
    """ Generic mkdocs-addresses exception. """


class NoStaticHandlerError(AddresserError):
    """ No StaticHandler instance declared yet (internal) """



class AddresserConfigError(AddresserError):
    """ Plugin config related error """


class InvalidOperationError(AddresserConfigError):
    """ Don't to that... (ConfigDescriptor specific => internal) """



#------------------------------------------------



class AbstractAddressError(AddresserError):
    """ Path/address related errors, that cover errors found when an address/path is
        found in a `<a href >` or `<img src >` tag.
    """


class MarkdownLinkError(AbstractAddressError):
    """ Addresses should never be written using the path of the md file, when using the plugin.
        Identifiers should be used instead.
    """

class NonAbsoluteAddressError(AbstractAddressError):
    """ Address should be absolute """

class NonRelativeAddressError(AbstractAddressError):
    """ Address is not considered relative (no leading dot). If this error occur on something
        that should be an absolute/external link, this means you probably should
        [mark it as ignored](--mkdocs_addresses_config_plugin_PluginOptions_ignored_identifiers_or_addresses) it.
    """

class LeadingSlashAddressError(AbstractAddressError):
    """ Address with leading slash are not allowed """

class TrailingSlashAddressError(AbstractAddressError):
    """ Address with trailing slash """

class NotAnImageAddressError(AbstractAddressError):
    """ The address isn't considered the one of an image. """

class NotAFileAddressError(AbstractAddressError):
    """ The corresponding file does not exist.

        (note: not using FileNotFoundError, so that all the errors are inheriting from the
        root class `AddresserError`)
    """

class NoMoveUpDirOnNonIndexPageUrlsAsDirAddressError(AbstractAddressError):
    """ When `use_directory_urls=true`, the address has to start with a "../" when defined in any
        markdown document that is not a `index.md` file (Except if the address is a bare anchor
        defined in the same page).
    """





class AnchorAddressError(AbstractAddressError):
    """ Invalid anchor found in an Address """


class ImageWithAnchorError(AnchorAddressError):
    """ Images src attribute should never contain an anchor """

class TooManyAnchorsError(AnchorAddressError):
    """ More than one `#` in the address """

class AnchorUnknownError(AnchorAddressError):
    """ Anchor not defined anywhere """

class AnchorNotInPageError(AnchorAddressError):
    """ The address is a bare anchor, but it is not found in the current page """

class EmptyAnchorError(AnchorAddressError):
    """ Address with a trailing anchor, but without identifier after it.
    <br>This is found with this kind of typo in the markdown document: `{: # wrong }`,
    which results in addresses like `".../#"`.
    """



#------------------------------------------------



class IdentifierError(AddresserError):
    """ Generic error related to identifiers: covers identifiers definitions and usages. """



class InvalidIdentifierError(IdentifierError):
    """ A reference identifier is invalid """

class DuplicateIdentifierError(IdentifierError):
    """ An identifier has already been registered """

class UnknownIdentifierError(IdentifierError):
    """ Unknown identifier """

class UnknownExternalLinkIdentifierError(IdentifierError):
    """ External link syntax that with an identifier which is not found in the
        [external_links_file](--mkdocs_addresses_config_plugin_PluginOptions_external_links_file)
    """



class OutdatedReferenceError(IdentifierError):
    """ Identifier used in some document, but that isn't defined anywhere anymore. """



#-------------------------------------------------

class ConcurrentPluginError(Exception):
    """ Using only for testing purpose, ensuring that pytest doesn't run in // """





from mkdocs_addresses.auto_completion_handler import AutoCompletion, RefKind
from mkdocs_addresses.soup_excluding_codes import PluginSoupExcludingCodes
from mkdocs_addresses.static_handler.tracker_references import References





def gen_invalid_external_links_when_autorefs_used(
    soup:PluginSoupExcludingCodes,
    defined_refs:References,
):
    """
    The autorefs plugin is modifying the html of external links ()"[...][...]") into span tags:

             <span data-autorefs-identifier="identifier" ...>

    These are replacing the original markdown that the plugin expects to find in the html when
    a reference doesn't exist, while the others 'with existing ref) have already been replaced

    BUT these spans also appear in the html generated by mkdocstrings, and will be replaced with
    links only during a post processing step (I didn't search when exactly, but in on_page_context
    hooks, they are still present as <span>, and not <a>). So the invalid references must be searched
    for very carefully.
    """
    external_links = AutoCompletion.get_for(RefKind.Ext)
    links          = AutoCompletion.get_for(RefKind.Link)
    for span in soup.find_all('span'):
        if not span.has_attr("data-autorefs-identifier"):
            continue
        identifier = span['data-autorefs-identifier']
        link_id    = links.get_final_identifier(identifier)
        ext_id     = external_links.get_final_identifier(identifier)
        if not defined_refs.has_id(link_id) and not defined_refs.has_id(ext_id):
            # if neither an external ref nor a header generated by mkdocstrings, this is invalid
            yield identifier



import os
import json
from textwrap import dedent
from dataclasses import dataclass
from typing import ClassVar, Dict, TYPE_CHECKING

from ._base_handler import AutoCompletion, RefKind, OPEN, CLOSE, SNIPPETS_TARGET, PLUGIN_MARK


if TYPE_CHECKING:
    from mkdocs_addresses.addresses_plugin import AddressAddressesPlugin






@dataclass
class VscAutoCompletion(AutoCompletion):

    filename: ClassVar[str] = '.vscode/links.code-snippets'


    @classmethod
    def store_other_snippets(cls, data:str):
        """ Extract some possible snippets unrelated to the plugin, if any, receiving the data as
            a bare string.
        """
        json_as_dict = json.loads(data).items()
        cls.other_snippets.extend(
            cls.build_other_snippet(name, d)
                for name,d in json_as_dict
                if 'description' not in d or PLUGIN_MARK not in d['description']
        )


    @classmethod
    def build_snippets_code(cls, plugin:'AddressAddressesPlugin'=None, _=True):
        if plugin is not None:
            data = (*cls.other_snippets, *plugin.global_handler.references.snippets.values())
        else:
            data = cls.other_snippets
        code = "{\n" + ",\n".join(sorted(data)) + "\n}"
        return code


    @classmethod
    def build_other_snippet(cls, name:str, dct:Dict):
        item = f'"{name}": { json.dumps(dct, indent=4) }'
        return item





@dataclass
class VscAutoAddress(VscAutoCompletion):
    """ Build snippets for links, ids, extras, files, images """


    def get_final_identifier(self, identifier:str, with_head=True):

        assets = 'assets/'
        if identifier.startswith(assets):
            identifier = identifier[len(assets):]
        return with_head * self.head + identifier.replace('.','_')


    def build_snippet(self, identifier, src_link, **_):

        clean_id = self.get_final_identifier(identifier, False)
        prefix   = f"{ self.kind.name }.{ clean_id }"       # Img.identifier or so...
        body     = self.__build_body(clean_id)
        yield self._build_snippet(prefix, body, src_link, "Md")

        if self.with_short:
            prefix = self.head + clean_id                   # !!identifier or so...
            body   = f"${ OPEN }0:{ self.head }{ CLOSE }{ clean_id }"
            yield self._build_snippet(prefix, body, src_link, "Reference")


    def __build_body(self, clean_id:str):
        """ helper... """
        L,R     = self.tail_wrap
        content = self._get_content(clean_id)
        body    = self.snippet_start + "[${0:" + content + "}]" + f"{ L }{ self.head }{ clean_id }{ R }"
        if self.attrs:
            body += "{: " + self.attrs + " }"
        return body


    def _get_content(self, clean_id:str):
        if self.kind == RefKind.Ext:
            return clean_id
        return 'content'


    def _build_snippet(self, prefix:str, body:str, src_link:str, name_prefix=''):
        if name_prefix:
            name_prefix += " - "
        json_id = f"{ name_prefix }{ self.kind.name }: { src_link !r}"
        return json_id, dedent(f"""\
            "{ json_id }": { OPEN }
                "prefix": "{ prefix }",
                "scope": { SNIPPETS_TARGET },
                "body": ["{ body }"],
                "description": "{ PLUGIN_MARK }"
            { CLOSE }""")






@dataclass
class VscAutoInclude(VscAutoAddress):
    """ Build snippets for markdown code inclusions (ie: "--8<--" ) """


    # pylint: disable-next=arguments-differ
    def build_snippet(self, identifier:str, src_link:str, *, root_inclusion:str): #, inclusions_with_root:bool):
        # root_inclusion: one of the path s given in plugin.inclusions (as uri)

        #i_truncate = len(root_inclusion)+(not inclusions_with_root)   # Extra slash to account for
        i_truncate = len(root_inclusion) + 1   # extra slash to account for
        _, ext     = os.path.splitext(identifier)
        short_uri  = identifier[i_truncate:]
        body       = f'--8<-- \\"{ identifier }\\"'
        prefix     = f"{ self.head }{ ext.strip('.') } { short_uri }"
        snippet    = self._build_snippet(prefix, body, identifier)
        yield snippet






from typing import ClassVar, Dict, TYPE_CHECKING
from dataclasses import dataclass

from mkdocs_addresses.toolbox import plugin_dump_padder
from ._vsc_handlers import VscAutoCompletion, VscAutoAddress, VscAutoInclude

if TYPE_CHECKING:
    from mkdocs_addresses.addresses_plugin import AddressAddressesPlugin




@dataclass
class NoAutoCompletion(VscAutoCompletion):
    """ Keep building the snippets/logic, but  """

    filename: ClassVar[str] = 'addresses_identifiers.txt'



    @classmethod
    def build_snippets_code(cls, plugin:'AddressAddressesPlugin'=None, as_json=False):
        if plugin is not None:
            data = plugin.global_handler.references.gen_refs_items()
        else:
            data = ()
        if as_json:
            code = "{" + "".join( f"{ plugin_dump_padder(identifier+':') } { file } \n"
                                    for identifier,file in sorted(data) ) + '}'
        else:
            code = "Identifier | Target\n--------------------\n\n" \
                + "".join( f"{ plugin_dump_padder(identifier+':') } { file } \n"
                            for identifier,file in sorted(data) )
        return code





@dataclass
class NoAutoAddress(VscAutoAddress): pass

@dataclass
class NoAutoInclude(VscAutoInclude): pass



import re

from typing import ClassVar, Dict, List, Tuple, TYPE_CHECKING
from enum import auto, IntEnum
from dataclasses import dataclass, field


if TYPE_CHECKING:
    from mkdocs_addresses.addresses_plugin import AddressAddressesPlugin







PLUGIN_MARK = "VSC-REF"

OPEN, CLOSE = '{', '}'

SNIPPETS_TARGET = '"markdown"'



class RefKind(IntEnum):
    Link = auto()
    File = auto()
    Img  = auto()
    Ext  = auto()
    Include = auto()





@dataclass
class AutoCompletion:
    """
    An AutoCompletion instance holds the whole logic to build individual code snippets as
    strings, for a given kind of data.
    """

    kind: RefKind               # Snippet type
    head: str                   # Prefix identifier (used to spot them in the html code)

    snippet_start: str = ''     # Element coming before the "[...]" initial part
    tail_wrap: Tuple[str,str] = '(',')'     # open+close for the last section (after "[...]")
    with_short: bool = True     # Also generate the short snippet version (head+identifier only)
    attrs: str = ""             # string of attributes to always add at the end of the body


    filename: ClassVar[str]     # Default dump file location (CwdPath)

    other_snippets: ClassVar[List[str]] = []
                                # Store all existing snippet unrelated to the plugin (if any)


    @classmethod
    def setup_handlers(cls, handlers:Dict[RefKind,'AutoCompletion']):
        AutoCompletion.__LINK_BUILDER          = handlers
        AutoCompletion.__POTENTIAL_REF_PATTERN = re.compile(
            '|'.join( re.escape(o.head) for o in AutoCompletion.__LINK_BUILDER.values() if o.head)
        )


    @classmethod
    def href_is_possible_ref(cls, href):
        return cls.__POTENTIAL_REF_PATTERN.match(href)


    @classmethod
    def get_for(cls, kind:RefKind) -> 'AutoCompletion':
        return cls.__LINK_BUILDER[kind]


    @classmethod
    def clear_other_snippets(cls):
        """ Clear the content of the snippets unrelated to the plugin """
        cls.other_snippets.clear()


    @classmethod
    def store_other_snippets(cls, data:str):
        """ Extract some possible snippets unrelated to the plugin, if any, receiving the data as
            a bare string.
        """
        raise NotImplementedError("Subclasses must override this method")


    @classmethod
    def build_snippets_code(cls, plugin:'AddressAddressesPlugin'=None, as_json=True):
        """ Rebuild the full snippets/identifiers data as a string, combining the "other_snippets"
            with the data coming from the References.
        """
        raise NotImplementedError("Subclasses must override this method")


    @classmethod
    def build_other_snippet(cls, name:str, dct:Dict):
        """ Build one snippet string for the snippets unrelated to the plugin """
        raise NotImplementedError("Subclasses must override this method")


    #------------------------------------------------------------------


    def build_snippet(self, identifier, src_link, **_):
        """ Generate code snippet entries for the given (bare) identifier, pointing toward the
            given source (which will, in the end, be the targeted element).
            @identifier: the bare identifier, without the "head" part ("!!", "--", ...)
            @src_link is the path of the source, relative to the cwd ("development wise")
            Automatically add the "shorthand" version of the snippet, if the kind requires it.
        """
        raise NotImplementedError("Subclasses must override this method")


    def get_final_identifier(self, identifier:str, with_head=True):
        """ Add the "head_link" part to the identifier if needed, and format the identifier
            to avoid they generate warnings during the builds
        """
        return with_head * self.head + identifier





from argparse import Namespace
import os
from pathlib import Path
import re

from typing import Any, Callable, List, Literal, Optional, Set, Tuple, TYPE_CHECKING, Type, Union

from mkdocs.config import config_options  as C
from mkdocs.config.base import Config
from mkdocs.config.defaults import MkDocsConfig

from . import path_manager
from .logger import logger
from .exceptions import AddresserConfigError, InvalidOperationError
from .auto_completion_handler import (
    AutoCompletion,
    IDE_OPTIONS, VSC, IdeKind,
    validate_config_and_get_auto_completion_handler,
)


if TYPE_CHECKING:
    from .addresses_plugin import AddressAddressesPlugin






# Definition used for mkdocstrings
class PluginOptions(Namespace):


    activate_cache: bool = True
    """
    Exploring the resulting html content of a page to gather and cross-check all the identifiers
    and addresses can be rather time-consuming. This means the plugin can significantly slow down
    the serve operations when the number of documents or their size become increase.

    To mitigate the issue, the plugin utilizes a cache (activated by default).

    When activated, only the recently modified markdown files will be re-evaluated during the next
    serve operation. This drastically reduces the serving time without changing the underlying
    logic of the plugin.

    Note that when mkdocs starts, all pages must be explored at least once, so the first serve
    operation can be very slow on big documentations.
    """


    black_list_pattern: str = "https?://|ftps?://|file:///|www\\."
    """
    Regex string: all addresses or identifiers that will
    [`re.match`](https://docs.python.org/3/library/re.html#re.match){: target=_blank } with it will be ignored.

    This allows to register absolute addresses, external links, or any other unusual scenarios you could
    think of, so that errors are not raised by the plugin when handling them.
    """

    black_listed_patterns: List[str] = []
    """
    Alternative to black_list_pattern, to get simpler declarations. Both will be merge with OR.
    """


    dump_action: str = 'normal'
    """
    The plugin normally dumps information about all the available references  on each serve
    (either as a file of code snippets, or as a text file. See
    [`dump_snippets_file`][mkdocs_addresses.config_plugin.PluginOptions.dump_snippets_file]
    option). The `dump_action` may change that behavior:

    * `normal`: the plugin recreates the informations on each serve.
    * `none`: the plugin behaves normally, but the
      [`dump_snippets_file`][mkdocs_addresses.config_plugin.PluginOptions.dump_snippets_file]
      is not updated.
    * `only`: the plugin gather the identifiers definitions, skips all verifications, dump the
      [`dump_snippets_file`][mkdocs_addresses.config_plugin.PluginOptions.dump_snippets_file]
      and finally raises [`DumpOnlyException`](--mkdocs_addresses_exceptions_DumpOnlyException)
      to stop serving.

        This option is useful when there's a need to regenerate code snippets while something
        in the addresses verifications steps causes a crash.
    """


    dump_snippets_file: str = ''
    """
    Define the location, relative to the cwd, of the file where the information regarding all
    the identifiers (either code snippets configuration or plain text) will be stored.

    There are restrictions about the name of this file:

    * The file cannot be in the `docs_dir` (it should never be anyway).
    * The extension of the file must be compatible with the value chosen for the
    [`ide`][mkdocs_addresses.config_plugin.PluginOptions.ide] option (see table below).
    * If left undefined, the filename is automatically defined, depending on the value of the
      [`ide`](--mkdocs_addresses_config_plugin_PluginOptions_ide) option:

    | `ide` | Default `dump_snippets_file` | Required file extension |
    |:-|:-|:-|
    | `none` | `addresses_identifiers.txt` | `.txt` |
    | `vsc` | `.vscode/links.code-snippets` | `.code-snippets` |
    """


    external_links_file: str = ''
    """
    Location of the file where are defined the global links/references, if any:

    ```markdown
    [mkdocs]: https://www.mkdocs.org/ "mkdocs - home"
    [id]: address "tooltip"
    ...
    ```

    If this option is defined:

    - This file cannot be in the docs_dir itself (it should never be anyway).
    - All links defined in the file must be absolute. If not, the plugin will raise a
      [NonAbsoluteAddressError](--mkdocs_addresses_exceptions_NonAbsoluteAddressError).
      It is possible to bypass this restriction by marking the address as ignored, using
      [`black_list_pattern`](--mkdocs_addresses_config_plugin_PluginOptions_black_list_pattern) or
      [`ignored_identifiers_or_addresses`](--mkdocs_addresses_config_plugin_PluginOptions_ignored_identifiers_or_addresses))
      configuration options.
    - This file should be appended automatically to all files through the
      [`pymdownx.snippets`][PyMdown-snippets]{: target=_blank } markdown extension, directly from
      the `mkdocs.yml` file. For example:

        ```yaml
        markdown_extensions:
            - pymdownx.snippets:
                check_paths: true
                auto_append: ["docs_logistic/external-links.md"]

        plugins:
            - mkdocs-addresses:
                external_links_file: docs_logistic/external-links.md
        ```
    """


    fail_fast: bool = False
    """
    This option determines whether the errors raised by the plugin will completely stop the
    serving process or will only provide information in the console without causing interruptions.

    During the documentation creation process, it can be advantageous to set this option to false.
    This way, there is no need to restart the server each time the plugin finds an incorrect link.
    Instead, you can continue working and address errors one by one.

    If `fail_fast` is set to false, a summary of all the wrong addresses found will be displayed
    once all the pages have been checked. This is especially helpful when there are a lot of
    errors and the content of the console becomes overwhelming.

    The summary is formatted like this:

    ```
    WARNING -  mkdocs-addresses: running with fail_fast=false...
               Summary of the exceptions raised while managing the references in the pages:
                   {error name}:    {error count}
                     {file location}     {identifier triggering the error}
    ```

    In these summaries, if the same identifier triggers several times the same error on the same
    page, it is counted only once.

    Note that some errors will appear separately from these summaries. Specifically,
    [`OutDatedReferenceError`](--mkdocs_addresses_exceptions_OutdatedReferenceError), which is
    checked at a different point during the validation process.
    """


    ide: IdeKind = VSC
    """
    Controls the kind of IDE used, for the autocompletion features. Possible values are:

    * `none`: no autocompletion activated. A simple txt file with all the available identifiers will
      be generated (see [`dump_snippets_file`][mkdocs_addresses.config_plugin.PluginOptions.dump_snippets_file]).

    * In any other case, code snippets will be automatically generated and saved in the appropriate
      [`dump_snippets_file`][mkdocs_addresses.config_plugin.PluginOptions.dump_snippets_file].

    Note:
        Remember that this option determines the kind of file used for the
        [`dump_snippets_file`][mkdocs_addresses.config_plugin.PluginOptions.dump_snippets_file]
        option: if a filename is assigned to it, its extension has to match with the chosen value
        for `ide` (see the [`dump_snippets_file`][mkdocs_addresses.config_plugin.PluginOptions.dump_snippets_file]
        section above).
    """


    ignore_auto_headers: bool = True
    """
    Controls whether the "bare headers" will be taken into account when the plugin collects
    identifiers from the documentation.
    <br>Note that both options have their advantages and disadvantages. For more details, see the
    section about [handling automatic mkdocs headers](--identifiers-handling-headers).

    In essence:

    * `ignore_auto_headers=true` (recommended):

        The same titles can be used in different locations, but headers must be explicitly marked
        with ids for the plugin to gather them.

        ```markdown
        ## Content
        ... is ignored

        ## Content
        ... is also ignored, hence, no error raised

        ## Content {: #hey-there }
        ... is gathered as --hey-there

        ### Hall of fame! {: #hall-of-fame }
        This example would result in a false negative and be ignored,
        because the id would be considered identical to the content.
        ```


    * `ignore_auto_headers=false`:

        In this case, all headers must be unique or have an id attribute (or
        [configured as ignored](--auto-headers-false)), otherwise a
        [`DuplicateIdentifierError`](--mkdocs_addresses_exceptions_DuplicateIdentifierError)
        is raised.

        ```markdown
        ## Content
        ... is gathered

        #### Content
        ... this one would raises DuplicateIdentifierError

        ## Content {: #precedence-over-content }
        ...this title wouldn't raise, because its id will be used instead
        ```
    """


    ignored_classes: List[str] = []
    """
    If defined, each string represent an html class name that will be ignored when collecting
    the identifiers and checks the addresses. Example of definition:

    ```yaml
    plugins:
        - mkdocs-addresses:
            ignored_classes:        # or: ignored_classes: ["class-name-1", "non-gathered"]
                - class-name-1
                - non-gathered
    ```

    !!! tip
        The class "headerlink" will systematically be excluded. This allows the plugin to ignore
        the permalinks generated by the [`toc`][toc]{: target=_blank } markdown extension.
    """

    skip_identifiers_definitions_in_pages: List[str] = []
    """
    List of uri strings, relative to the docs, of pages where identifiers definitions won't
    be gathered.

    May also be a `#!py Path.glob` uri pattern. Patterns are identified as such when the
    `#!py "*"` character is used in them.

    Constraints:
    - Patterns have to be written relative to the docs_dir
    - Patterns are not recursive by default, but can be when using `#!py "**"` in them (keep in
    mind this means "0 or more", with `Path.glob`)
    - Patterns have to specifically target the md file, so they have to end with `#!py "*.md"`.
    """


    ignored_identifiers_or_addresses: List[str] = []
    """
    If defined, each string represents an html id or a plugin identifier or an address (`<a href>`,
    or `<img src>` attributes) that will be ignored when the plugin collects and checks identifiers
    or addresses.
    Example of definition:

    ```yaml
    plugins:
        - mkdocs-addresses:
            ignored_identifiers_or_addresses:
                - subtitle
                - skip-this-id
                - --skip-this-also
                - ../meta/invalid/skipped.anyway
    ```

    !!! tip
        Ids starting with double underscores will always be ignored, because some plugins create
        and utilize them to mark some tags for their internal logic.
    """


    imgs_extension_pattern: str = "(?i)\\.(jpg|jpeg|bmp|png|svg)$"
    """
    Regex pattern used to identify files that are images, using
    [`re.search`](https://docs.python.org/3/library/re.html#re.search){: target=_blank }.
    """


    inclusions: List[str] = []
    """
    Slash separated paths, relative to the cwd, that point to directories containing files
    to be included using the "scissors" operator: `--<8--` (requires the `pymdownx.snippets`
    markdown extension).

    More information about the resulting code snippets in the [autocompletion section](--inclusions-snippets).

    !!! note
        The directories specified under `inclusions` are not supposed to be in the `docs_dir`.
        If they are, mkdocs will generate warnings about files not included in the nav.
    """


    # inclusions_with_root: bool = False
    # """
    # Define if the parent inclusion directory will be part of the code snippet identifier.

    # This is to prevent potential conflicts of  snippets identifiers when several inclusions
    # directories are used. See details in the [autocompletion section](--inclusions-snippets).
    # """


    more_verbose: bool = False
    """
    Controls the verbosity level of the plugin. It set to `true`, the verbosity is slightly
    increased, without affecting the level of verbosity of the mkdocs logger.

    If mkdocs is run in verbose mode (using `mkdocs serve -v`), this option is ignored and
    the verbosity of the plugin's logger is also set to DEBUG level.
    """


    plugin_off: bool = False
    """
    If true, completely deactivates the plugin.

    This may speed up the serve steps, but the whole verification logic of the plugin will be lost.
    """


    strict_anchor_check: bool = True
    """
    During anchors validations, when `strict_anchor_check` is set to true, an anchor has to match
    an identifier to be considered valid.
    If set to false, the plugin will no longer raise errors regarding unknown anchors. Instead,
    it will display a warning in the console.

    !!!note "Rationale"
        Users are encouraged to use the plugin's identifiers but:

        1. They could use bare anchors to write links within the current document.
        1. They could have added the plugin to an old project which utilizes the "old-fashion"
           addresses (`../a/b/c.md`).

        By essence, neither of these situations is problematic but, when
        [`ignore_auto_headers`][mkdocs_addresses.config_plugin.PluginOptions.ignore_auto_headers]
        is set to true (which is the default value...), the plugin cannot know anymore about the
        automatic headers and will assume the anchor is wrong if it doesn't match any of its
        registered identifiers.

        This behavior can be turned off by setting `strict_anchor_check` to false. However, keep
        in mind this is discouraged because doing so significantly reduces the level of
        assurance the plugin can provide regarding the correctness of all the addresses within
        the documentation.

        In summary:

        * Setting `strict_anchor_check` to false is not recommended, as it diminishes the
          reliability of the plugin's verifications.
        * But it might be useful to facilitate the integration of the plugin into older projects.
    """

    # LEGACY:
    use_vsc: Union[bool,None] = None
    # """
    # Controls the kind of file the plugin will generate with all the collected identifiers.

    # * If true, a VSC `.code-snippets` file will be created, enabling autocompletion features.
    #   This file is located at the value specified for the
    #   [`dump_snippets_file`][mkdocs_addresses.config_plugin.PluginOptions.dump_snippets_file]
    #   option.
    # * If false, the plugin overall behavior remains unchanged, but a simple `.txt` file is
    #   generated instead, containing the identifiers and the related targeted address.

    # Note:
    #     Remember that this option determines the kind of file used for the
    #     [`dump_snippets_file`][mkdocs_addresses.config_plugin.PluginOptions.dump_snippets_file]
    #     option. This means that if you assigned a filename to that option, its extension has
    #     to match with the chosen value for `use_vsc` (see the
    #     [`dump_snippets_file`][mkdocs_addresses.config_plugin.PluginOptions.dump_snippets_file]
    #     section above).
    # """


    verify_only: List[str] = []
    """
    Each entry denotes the location of the only files (slash separated, and relative to the
    _^^`docs_dir`^^_) that will be validated by the plugin. All other operations proceed as usual.

    If the list remains empty, the plugin validates all markdown files within the `docs_dir`.

     For example, to solely validate the `docs/subdir/blabla.md` file, use:
    ```yaml
    plugins:
        - mkdocs-addresses:
            verify_only:
                - subdir/blabla.md
    ```
    """


CONFIG_ANNOTATIONS = set(PluginOptions.__annotations__) - { "dump_file_vsc", "dump_file_txt" }






DUMP_OPTIONS = tuple('normal only none'.split())
NORMAL, DUMP_ONLY, NO_DUMP = DUMP_OPTIONS
assert PluginOptions.dump_action in DUMP_OPTIONS, "Invalid plugin setup"




# Config types definitions...:

ConfigErrors   = List[Tuple[str,bool,str,Optional[Exception]]]
ConfigWarnings = List[Tuple[str,str]]
FatalCbk       = Callable





class AddressAddressesConfig(Config):

    activate_cache = C.Type(bool, default=PluginOptions.activate_cache)

    dump_action = C.Choice(DUMP_OPTIONS, default=PluginOptions.dump_action)

    dump_snippets_file = C.Type(str, default='')      # Automatically overridden at validation time

    external_links_file = C.Type(str, default=PluginOptions.external_links_file)

    fail_fast = C.Type(bool, default=PluginOptions.fail_fast)

    ide = C.Choice(IDE_OPTIONS, default=PluginOptions.ide)

    imgs_extension_pattern = C.Type(str, default=PluginOptions.imgs_extension_pattern)

    ignore_auto_headers = C.Type(bool, default=PluginOptions.ignore_auto_headers)

    skip_identifiers_definitions_in_pages = C.ListOfItems(C.Type(str), default=PluginOptions.skip_identifiers_definitions_in_pages)

    ignored_identifiers_or_addresses = C.ListOfItems(C.Type(str), default=PluginOptions.ignored_identifiers_or_addresses)

    ignored_classes = C.ListOfItems(C.Type(str), default=PluginOptions.ignored_classes)

    inclusions = C.ListOfItems(C.Type(str), default=PluginOptions.inclusions)

    # inclusions_with_root = C.Type(bool, default=PluginOptions.inclusions_with_root)

    black_list_pattern = C.Type(str, default=PluginOptions.black_list_pattern)

    black_listed_patterns = C.ListOfItems(C.Type(str), default=PluginOptions.black_listed_patterns)

    more_verbose = C.Type(bool, default=PluginOptions.more_verbose)

    plugin_off = C.Type(bool, default=PluginOptions.plugin_off)

    strict_anchor_check = C.Type(bool, default=PluginOptions.strict_anchor_check)

    # LEGACY:
    use_vsc = C.Optional(C.Choice((True,False,None), default=PluginOptions.use_vsc))

    verify_only = C.ListOfItems(C.File(exists=True), default=PluginOptions.verify_only)

#TypeError: AddressAddressesConfig.use_vsc: Setting 'required' is unsupported in class-based configs. All values are required, or can be wrapped into config_options.Optional

    #----------------------------------------------
    def _validate(self):
        return super()._validate()

    def _pre_validate(self):
        return super()._pre_validate()

    def _post_validate(self):
        return super()._post_validate()


    def validate_and_process(self, config:MkDocsConfig):
        """ Performs the basic, then extras validations on each property, and also apply the
            post conversions to the data structures that need it.
            Add on the fly the needed fields from the mkdocs config.
        """
        # Enforce no unexpected option name (because this is boring AF...)
        if 'plugins' in config and 'mkdocs-addresses' in config.plugins:
            settings = set(config.plugins['mkdocs-addresses'].config)
            base_settings = set(dir(config.plugins['mkdocs-addresses'].__class__.mro()[1]))
            user_settings = settings - base_settings
        else:
            # when testing in mkdocs-addresses, the plugin isn't in the yml file (must change
            # that: see use of entry-points in pyodide-mkdocs-theme):
            user_settings = set(self.user_configs[-1])

        nope = user_settings - CONFIG_ANNOTATIONS
        if nope:
            nope    = ''.join( '\n   '+key for key in sorted(nope))
            options = ''.join( '\n   '+key for key in sorted(CONFIG_ANNOTATIONS))
            raise AddresserConfigError(
                f"Unexpected option(s) name(s):{ nope }\n\n Available options are:{options}"
            )

        # Extract completion handler class, handling use_vsc, ide and dump_snippets_file defaults
        # (no validation):
        completion_class = validate_config_and_get_auto_completion_handler(self)


        # Perform "~~post"~~" PRE-conversions:       (yeah, "post"... lol...)
        #-------------------------------------

        maybe_abs = path_manager.to_os(config['docs_dir'])
        self.uni_docs_dir = maybe_abs.absolute().relative_to(Path.cwd())

        for prop in 'docs_dir  use_directory_urls'.split():
            setattr(self, prop, config[prop])

        self.black_list_pattern = "|".join(
            [self.black_list_pattern] + self.black_listed_patterns
        ).strip('|')

        post_conversions: List[Tuple[Callable[[str],Any], str]] = [
            (re.compile, 'imgs_extension_pattern  black_list_pattern'),
            (set,        'ignored_identifiers_or_addresses ignored_classes verify_only inclusions skip_identifiers_definitions_in_pages'),
        ]
        for conversion,props in post_conversions:
            for prop in props.split():
                value = getattr(self, prop)
                try:
                    post = conversion(value)
                    setattr(self, prop, post)
                except Exception as e:                              # pylint: disable=all
                    e.args = ( f"{ prop }: { e.args[0] }",)
                    raise

        self.ignored_classes.add('headerlink')      # this avoids to consider mkdocs permalinks in headers, when used

        # More specific validations and post conversions:
        #------------------------------------------------

        self._validate_inclusions_dir()
        self._validate_black_list_pattern()
        self._validate_external_links_file()
        self._validate_verify_only_md_files()
        self._validate_verify_only_md_files()
        self._resolve_patterns_and_validate_skip_identifiers_definitions_in_pages()
        self._validate_auto_completion_config(completion_class)
        self._validate_dump_extension_vs_ide(completion_class)

        return completion_class


    #----------------------------------------------


    def _check_not_in_docs(self, prop:str, path:Union[str,Path]):
        rel = os.path.relpath(path, self.uni_docs_dir)
        if not rel.startswith('..'):
            raise AddresserConfigError(
                f"{prop}: {path!r} should not be inside the docs_dir directory"
            )


    #----------------------------------------------


    def _validate_inclusions_dir(self):
        self.inclusions = sorted(map( path_manager.to_os, self.inclusions))
        oops = [location for location in self.inclusions if not location.is_dir() ]
        if oops:
            raise AddresserConfigError(
                f"'inclusions' should be a list of directories, but found invalid paths: ({ oops !r})"
            )

        for path in self.inclusions:
            self._check_not_in_docs('inclusions', path)



    def _validate_black_list_pattern(self):
        if not self.black_list_pattern.pattern:      # this is now a pattern, not a string anymore!
            raise AddresserConfigError(
                "black_list_pattern should never be an empty string. Note that not using at "
                "least the default pattern will definitely cause problems..."
            )



    def _validate_external_links_file(self):
        if not self.external_links_file:
            return

        self.external_links_file = p = path_manager.to_os(self.external_links_file)
        if not p.is_file() or not p.suffix=='.md':
            raise AddresserConfigError(
                "'external_links_file' is configured, but is not an existing markdown file "
                f'("{ self.external_links_file }")'
            )
        self._check_not_in_docs('external_links_file', self.external_links_file)



    def _validate_auto_completion_config(self, completion_class:Type[AutoCompletion]):
        self.dump_snippets_file = path_manager.to_os(self.dump_snippets_file)
        self._check_not_in_docs('dump_snippets_file', self.dump_snippets_file)


    def _validate_dump_extension_vs_ide(self, completion_class:Type[AutoCompletion]):
        # Note: separated from _validate_auto_completion_config for testing purpose.
        needed_suffix:str = path_manager.to_os(completion_class.filename).suffix
        if self.dump_snippets_file.suffix != needed_suffix:
            raise AddresserConfigError(
                "Inconsistent configuration:\n"
                f"   ide: {self.ide}\n"
                f"   dump_snippets_file: {self.dump_snippets_file}"
            )



    def _validate_verify_only_md_files(self):
        invalids = [ file for file in self.verify_only
                          if not path_manager.to_os(file).suffix=='.md']
        if invalids:
            raise AddresserConfigError(
                "All options for `verify_only` should be valid markdown files, relative to the cwd.\n"
                +"Invalid entries:" + "".join('\n    '+file for file in invalids)
            )

    def _resolve_patterns_and_validate_skip_identifiers_definitions_in_pages(self):

        # Handle patterns:
        patterns = {p for p in self.skip_identifiers_definitions_in_pages if "*" in p}
        self.skip_identifiers_definitions_in_pages -= patterns

        # Validate all files:
        invalids = [
            file for file in map(Path, self.skip_identifiers_definitions_in_pages)
                 if not ('docs' / file).is_file()
        ]
        if invalids:
            raise AddresserConfigError(
                "The following files do not exist, relative to the docs_dir:"
                +"".join(f'\n    { file.as_posix() }' for file in invalids)
            )
        # Validate patterns:
        invalid_patterns = [
            p for p in patterns if not p.endswith("*.md")
        ]
        if invalid_patterns:
            raise AddresserConfigError(
                "The following patterns are not valid: patterns have to end with '*.md'."
                +"".join(f'\n    {p}' for p in invalid_patterns)
            )

        logger.info('Pages exclusions patterns applied:' + ''.join(f"\n    - {p!r}" for p in patterns))

        # Add md files found from pattern matching (non recursive by default. Use "**" if needed)
        from_patterns: List[str] = []
        docs = Path('docs')
        for pattern in patterns:
            for file in docs.glob(pattern):
                if file.suffix == '.md':
                    as_posix = file.as_posix()
                    from_patterns.append(as_posix)

        # Rebuild the actual content needed (string uri, without "docs/" to match
        # against page.file.src_uri):
        skip_docs_slash = len('docs/')
        self.skip_identifiers_definitions_in_pages.update(
            file[skip_docs_slash:] for file in from_patterns
        )











class ConfigDescriptor:
    """ Dynamic relay data descriptor, to articulate the plugin and its configuration """

    prop: str

    def __set_name__(self, _, prop:str):
        self.prop = prop

    def __get__(self, obj:'AddressAddressesPlugin', _:type):
        return getattr(obj.config, self.prop)

    def __set__(self, obj:'AddressAddressesPlugin', value:Any):
        # setter needed to make this a data descriptor, but should never be used.
        raise InvalidOperationError(
            "Config properties should never be reassigned from the plugin instance."
        )








def __check_config_annotations_consistency():
    """
    Verification that no option of PluginOptions has been forgotten in AddressAddressesConfig:
    """
    implemented = set(AddressAddressesConfig.__annotations__) - {
        "_schema",                  # inherited
        "config_file_path",         # inherited
    }
    missing = implemented - CONFIG_ANNOTATIONS
    assert not missing, f"Not implemented: { ', '.join(missing) }"

__check_config_annotations_consistency()





from collections import defaultdict
from typing import Dict, List, Set

from mkdocs.structure.pages import Page
from mkdocs_addresses import path_manager

from mkdocs_addresses.static_handler.tracker_data_pages import PageDataTracker
from mkdocs_addresses.toolbox import plugin_dump_padder

from .types_aliases import PathCwd



JOINER = ' | '


class ContextExceptionsTracker(PageDataTracker):
    """ Keep track of the sources files (as uris) that are using references, like:

            ![alt](!!logo_svg)

        References can only be _used_ in pages, so the sources are the page.file.src_uri.
        Normally, n need for OSCwdPath thingy or so... BUT, to get a homogeneous behavior
        of the whole plugin, store the cwd rooted paths instead.
    """


    def archive_current(self, uni_docs_dir:PathCwd):
        """ Dump a deepcopy of the to internal structures: (id_to_sources, source_to_ids) """
        super().archive_current(uni_docs_dir)


    def add_exc(self, exc_name:str, ref:str, page:Page):
        key = f'{ exc_name }{ JOINER }{ ref }'
        self.add_id(page, key)


    def get_errors_info(self):
        """ Build an error message listing all the errors that are still registered
        """
        # rebuild data "per exception kind":
        dct = defaultdict(list)
        for k,files in self.id_to_sources.items():
            exc_name,ref = k.split(JOINER)
            for cwd_path in files:
                uri = path_manager.to_uri(cwd_path)
                dct[exc_name].append( (uri,ref) )

        sorted_items = sorted(dct.items(), key=lambda it: -len(it[1]))
        info_message = ''.join(
            self.__format_exc_and_info(exc_name, sorted(data)) for exc_name,data in sorted_items
        )
        return info_message


    @staticmethod
    def __format_exc_and_info(exc_name:str, lst_infos:List[str]):
        title = f'\n\033[31m{ PAD4 }{plugin_dump_padder(exc_name+":")} { len(lst_infos) } \033[0m'
        lines = "".join(f"\n{ PAD6 }{plugin_dump_padder(uri)} { ref }" for uri,ref in lst_infos)
        return title + lines



PAD4 = '    '
PAD6 = '      '
MIDDLE_MSG = "used in file"  # this is used in the tests to cut properly a line into the original information
